"""HTML Form handling for web clients.

ClientForm is a Python module for handling HTML forms on the client
side, useful for parsing HTML forms, filling them in and returning the
completed forms to the server.  It is a port (more-or-less) of Gisle
Aas' Perl module HTML::Form, from the libwww-perl library.

RFC 1866: HTML 2.0
HTML 4.01 Specification, W3C Recommendation 24 December 1999
RFC 1867: Form-based file upload


Copyright 1998-2000 Gisle Aas.
Copyright 2002 John J. Lee <jjl@pobox.com> (The Python port)

This code is free software; you can redistribute it and/or modify it
under the terms of the Perl Artistic License.

"""

# XXX
# Fix protocol issues.
# label attribute for all controls? (note that SELECT label defaults to SELECT
#  contents).  I think there is some variation between elements here -- if so,
#  this would be useful.
# Release stable version.
# Add FILE (FormParser, FileControl -- and HTMLForm?) -- for file upload.
#  multipart/form-data.
# Release beta & stable version with FILE.
# Wait a while for bugs to turn up.
# Write docs in Python TeX format.
# Submit as Python standard library candidate (somebody needs to bully Gisle
#  about the license).

# Protocol issues:
# Not sure whether or not relative URIs should be joined with base URI.
# Why on earth doesn't google take params implied by form?? (cftest.py)
#  Possibly user-agent, etc., but that doesn't explain why netscape sends
#  the peculiar URL that it does.
# Since ISINDEX has no name attribute, how does it get submitted??

import sys
if sys.version[:5] == "1.5.2":
    PYTHON_VERSION = 152
else:
    PYTHON_VERSION = reduce(lambda x, y: 10*x+y, sys.version_info[:3])
if PYTHON_VERSION < 230:
    True = 1
    False = 0

import htmllib, formatter, urllib, urllib2, types, string, sys
from types import StringType
try:
    import UnicodeType
except ImportError:
    UNICODE = False
else:
    UNICODE = True
from urlparse import urlparse, urljoin
from copy import copy

VERSION = "0.0.3b"  # based on Gisle Aas's version 0.03, libwww-perl 5.64

# This version of urlencode is from my Python 1.5.2 back-port of the
# Python 2.1 CVS maintenance branch of urllib.  It will accept a sequence
# of pairs instead of a mapping -- the 2.0 version only accepts a mapping.
def urlencode(query,doseq=0):
    """Encode a sequence of two-element tuples or dictionary into a URL query string.

    If any values in the query arg are sequences and doseq is true, each
    sequence element is converted to a separate parameter.

    If the query arg is a sequence of two-element tuples, the order of the
    parameters in the output will match the order of parameters in the
    input.
    """

    if hasattr(query,"items"):
        # mapping objects
        query = query.items()
    else:
        # it's a bother at times that strings and string-like objects are
        # sequences...
        try:
            # non-sequence items should not work with len()
            x = len(query)
            # non-empty strings will fail this
            if len(query) and type(query[0]) != types.TupleType:
                raise TypeError
            # zero-length sequences of all types will get here and succeed,
            # but that's a minor nit - since the original implementation
            # allowed empty dicts that type of behavior probably should be
            # preserved for consistency
        except TypeError:
            ty,va,tb = sys.exc_info()
            raise TypeError, "not a valid non-string sequence or mapping object", tb

    l = []
    if not doseq:
        # preserve old behavior
        for k, v in query:
            k = urllib.quote_plus(str(k))
            v = urllib.quote_plus(str(v))
            l.append(k + '=' + v)
    else:
        for k, v in query:
            k = urllib.quote_plus(str(k))
            if type(v) == types.StringType:
                v = urllib.quote_plus(v)
                l.append(k + '=' + v)
            elif UNICODE and type(v) == types.UnicodeType:
                # is there a reasonable way to convert to ASCII?
                # encode generates a string, but "replace" or "ignore"
                # lose information and "strict" can raise UnicodeError
                v = urllib.quote_plus(v.encode("ASCII","replace"))
                l.append(k + '=' + v)
            else:
                try:
                    # is this a sufficient test for sequence-ness?
                    x = len(v)
                except TypeError:
                    # not a sequence
                    v = urllib.quote_plus(str(v))
                    l.append(k + '=' + v)
                else:
                    # loop over the sequence
                    for elt in v:
                        l.append(k + '=' + urllib.quote_plus(str(elt)))
    return string.join(l, '&')


def issequence(object):
    """Test whether object is a sequence."""
    try:
        object[0]
    except TypeError:
        return False
    except KeyError:
        return False
    except IndexError:
        pass
    return True

def isstringlike(object):
    if (isinstance(object, StringType) or
        (UNICODE and isinstance(object, UnicodeType))):
        return True
    return False

def is_relative_URI(uri):
    return not (urlparse(uri)[2]).startswith("/")

class ParseError(Exception): pass

def ParseResponse(response, select_default=False):
    """Parse HTTP response and return a list of HTMLForm instances.

    The return value of urllib2.urlopen can be conveniently passed to this
    function as the response parameter.

    response: file-like object (supporting read() method) with a method
     geturl(), returning the base URI of the HTTP response
    select_default: for multiple-selection SELECT controls, pick the first item
     as the default if none are selected in the HTML

    Pass a true value for select_default if you want the behaviour specified by
    RFC 1866 (the HTML 2.0 standard), which is to select the first item in a
    multiple-select SELECT control if none were selected in the HTML.  Most
    browsers (including Microsoft Internet Explorer (IE) and Netscape
    Navigator) instead leave all items unselected in this case, and the W3C
    HTML 4.0 standard leaves this behaviour undefined.

    """
    return ParseFile(response, response.geturl())

def ParseFile(file, base_uri, select_default=False):
    """Parse HTML and return a list of HTMLForm instances.

    file: file-like object (supporting read() method) containing HTML with zero
     or more forms to be parsed
    base_uri: the base URI of the document
    select_default: for multiple-selection SELECT controls, pick the first item
     as the default if none are selected in the HTML

    Pass a true value for select_default if you want the behaviour specified by
    RFC 1866 (the HTML 2.0 standard), which is to select the first item in a
    multiple-select SELECT control if none were selected in the HTML.  Most
    browsers (including Microsoft Internet Explorer (IE) and Netscape
    Navigator) instead leave all items unselected in this case, and the W3C
    HTML 4.0 standard leaves this behaviour undefined.

    """
    fp = FormParser()
    fp.feed(file.read())
    forms = []
    for (name, action, method, enctype), attrs, controls in fp.forms:
        if action is None:
            action = base_uri
        # XXX Not sure whether or not relative URIs should be joined with base
        # URI.
##         elif is_relative_URI(action):
        else:
            action = urljoin(base_uri, action)
        form = HTMLForm(action, method, enctype, name, attrs)
        for type, attr in controls:
            form.new_control(type, attr, select_default=select_default)
        forms.append(form)
    for form in forms:
        form.fixup()
    return forms

class FormParser(htmllib.HTMLParser):
    # pinched (and modified) from Moshe Zadka
    def __init__(self):
        htmllib.HTMLParser.__init__(self, formatter.NullFormatter())
        self.forms = []
        self._current_form = None
        self._select = None
        self._option = None
        self._textarea = None

    def start_form(self, attrs):
        if self._current_form is not None:
            raise ParseError, "nested FORMs"
        name = None
        action = None
        enctype = "application/x-www-form-urlencoded"
        method = "GET"
        d = {}
        for key, value in attrs:
            if key == "name":
                name = value
            elif key == "action":
                action = value
            elif key == "method":
                method = string.upper(value)
            elif key == "enctype":
                enctype = value
            else:
                d["key"] = value
        controls = []
        self._current_form = (name, action, method, enctype), d, controls

    def end_form(self):
        if self._current_form is None:
            raise ParseError, "end of FORM before start"
        self.forms.append(self._current_form)
        self._current_form = None

    def start_select(self, attrs):
        if self._current_form is None:
            raise ParseError, "start of SELECT before start of FORM"
        if self._select is not None:
            raise ParseError, "nested SELECTs"
        if self._textarea is not None:
            raise ParseError, "SELECT inside TEXTAREA"
        d = {}
        for key, val in attrs:
            d[key] = val

        value = d.get("value")
        if value is not None:
            d["select_value"] = value

        self._select = d

    def end_select(self):
        if self._current_form is None:
            raise ParseError, "end of SELECT before start of FORM"
        if self._select is None:
            raise ParseError, "end of SELECT before start"
        self._select = None

    def do_option(self, attrs):
        if self._current_form is None:
            raise ParseError, "OPTION outside of FORM"
        if self._select is None:
            raise ParseError, "OPTION outside of SELECT"
        d = {}
        for key, val in attrs:
            d[key] = val

        self._option = self._select.copy()
        self._option.update(d)

    def start_textarea(self, attrs):
        if self._current_form is None:
            raise ParseError, "start of TEXTAREA before start of FORM"
        if self._textarea is not None:
            raise ParseError, "nested TEXTAREAs"
        if self._select is not None:
            raise ParseError, "TEXTAREA inside SELECT"
        d = {}
        for key, val in attrs:
            d[key] = val

        self._textarea = d

    def end_textarea(self):
        if self._current_form is None:
            raise ParseError, "end of TEXTAREA before start of FORM"
        if self._textarea is None:
            raise ParseError, "end of TEXTAREA before start"
        self._textarea = None

    def handle_data(self, data):
        if self._option is not None:
            contents = string.strip(data)
            controls = self._current_form[2]
            if not self._option.has_key("value"):
                self._option["value"] = contents
            controls.append(("select", self._option))
            self._option = None
        elif self._textarea is not None:
            contents = string.strip(data)
            controls = self._current_form[2]
            self._textarea["value"] = contents
            controls.append(("textarea", self._textarea))

    def do_button(self, attrs):
        if self._current_form is None:
            raise ParseError, "start of BUTTON before start of FORM"
        d = {}
        d["type"] = "submit"
        for key, val in attrs:
            d[key] = val
        controls = self._current_form[2]

        type = d["type"]
        del d["type"]
        controls.append((type, d))

    def do_input(self, attrs):
        if self._current_form is None:
            raise ParseError, "start of INPUT before start of FORM"
        d = {}
        d["type"] = "text"
        for key, val in attrs:
            d[key] = val
        controls = self._current_form[2]

        type = d["type"]
        del d["type"]
        controls.append((type, d))

    def do_isindex(self, attrs):
        if self._current_form is None:
            raise ParseError, "start of ISINDEX before start of FORM"
        d = {}
        d["type"] = "isindex"
        for key, val in attrs:
            d[key] = val
        controls = self._current_form[2]

        type = d["type"]
        del d["type"]
        controls.append((type, d))


class HTMLForm:
    """Represents a single HTML <form> ... </form> instance.

    A form consists of a sequence of controls that usually have names, and
    which can take on various values.

    An HTMLForm is only properly initialised after the fixup method has been
    called.  See ListControl.__doc__ for the reason this is required.  Usually,
    however, the constructor is not called directly; rather, the ParseFile or
    ParseResponse factory functions are used.  In that case, fixup is called
    for you:

    import ClientForm
    forms = ClientForm.ParseFile(html, base_uri)
    form = forms[0]
    form["query"] = "Python"
    response = urllib2.urlopen(form.click())

    Public attributes:

    action: full (absolute URI) form action
    method: "GET" or "POST"
    enctype: form transfer encoding MIME type
    name: name of form (None if no name was specified)
    attrs: dictionary mapping other HTML form attributes to their values

    controls: list of Control instances; do not append to this list
     (instead, call form.new_control to make a Control and add it to the
     form, or control.add_to_form if you already have a Control instance)

    """
    def __init__(self, action, method="GET",
                 enctype="application/x-www-form-urlencoded",
                 name=None, attrs=None):
        """
        In the usual case, use ParseResponse (or ParseFile) to create new
        HTMLForm objects.

        action: full (absolute URI) form action
        method: "GET" or "POST"
        enctype: form transfer encoding MIME type
        name: name of form
        attrs: dictionary mapping all other HTML form attributes to their
         values

        """
        self.action = action
        self.method = method
        self.enctype = enctype
        self.name = name
        if attrs is not None:
            self.attrs = attrs.copy()
        else:
            self.attrs = {}
        self.controls = []

    def new_control(self, type, attrs,
                    ignore_unknown=False, select_default=False):
        """Adds a new control to the form.

        Note that controls representing lists of items are built up from
        controls holding only a single item.  See ListControl.__doc__ for
        further information.

        type: type of control (see Control.__doc__ for a list)
        attrs: HTML attributes of control
        ignore_unknown: if true, use a dummy Control instance for controls of
         unknown type; otherwise, raise ValueError
        select_default: for multiple-selection SELECT controls, pick the first
         item as the default if no 'selected' HTML attribute is present (this
         defaulting happens when the HTMLForm.fixup method is called)

        """
        type = string.lower(type)
        klass = TYPE2CLASS.get(type)
        if klass is None:
            if ignore_unknown:
                klass = IgnoreControl
            else:
                raise ValueError, "Unknown control type '%s'" % type

        a = attrs.copy()
        if issubclass(klass, ListControl):
            control = klass(type, a, select_default)
        else:
            control = klass(type, a)
        control.add_to_form(self)

    def fixup(self):
        """Normalise form after all controls have been added.

        This method should only be called once, after all controls have been
        added to the form.

        """
        for control in self.controls:
            control.fixup()

    def find_control(self, name=None, type=None, nr=0):
        """Locate some specific control within the form.

        At least one of the arguments must be defined.  If no matching control
        is found, None is returned.

        If name is specified, then the control must have the indicated name.
        If type is specified then the control must have the specified type.  In
        addition to the types possible for <input> HTML tags, we also have
        "button", "textarea" and "select".  nr is the sequence number of the
        control with the indicated name and/or type (where 0 is the first).
        Note that control 0 is the first control matching the name and type
        arguments (if supplied); it is not the first control in the form.

        """
        if (name is None) and (type is None) and (nr is None):
            raise ValueError, \
                  "At least one of name, type and nr args must be supplied"
        if nr < 0:
            raise ValueError, "nr must be a positive integer"
        for control in self.controls:
            if name is not None:
                controlname = control.name
                if controlname is None:
                    continue
                if name != controlname:
                    continue
            if type is not None and type != control.type:
                continue
            if nr:
                nr = nr - 1
                continue
            return control
        return None

    def possible_values(self, name=None, type=None, nr=0):
        """Return a list of values for the specified control.

        See find_control.__doc__ for the meanings of the arguments.

        For controls that do not have a prescribed set of values, this returns
        an empty list.

        """
        ipt = self.find_control(name, type, nr)
        return ipt.possible_values()

    def toggle(self, name, value):
        """Toggle a named ListControl's value."""
        control = self.find_control(name)
        if control is None:
            raise ValueError, "control %s not found in form" % name
        if not hasattr(control, "toggle"):
            raise TypeError, \
                  "control %s does not have toggle-able values" % name
        control.toggle(value)

    def __setitem__(self, name, value):
        """Set the value(s) of named control."""
        control = self.find_control(name)
        try:
            control.value = value
        except AttributeError, e:
            raise ValueError, str(e)

    def __getitem__(self, name):
        """Get the value(s) of named control."""
        control = self.find_control(name)
        return control.value

    def items(self):
        """Return sequence of (key, value) pairs suitable for urlencoding.

        You probably want to use the click method instead, possibly with the
        items argument.

        Note that Python 2.0's urllib.urlencode is slightly broken: it only
        accepts a mapping, not a sequence of pairs, as an argument.  This
        messes up any ordering in the argument.  Use ClientForm.urlencode
        instead.

        """
        items = []
        for control in self.controls:
            items.extend(control.items())
        return items

    def make_request(self):
        """Return a urllib2.request object reflecting current form settings.

        You probably want to use the click method instead.

        Currently, ValueError is raised if the HTMLForm has method "POST" and
        an enctype that is not "application/x-www-form-urlencoded".

        """
        method = string.upper(self.method)
        if method == "GET":
            uri = "%s?%s" % (self.action, urlencode(self.items()))
            return urllib2.Request(uri)
        elif method == "POST":
            if self.enctype != "application/x-www-form-urlencoded":
                raise ValueError, \
                      "unknown form encoding type %s" % self.enctype
            req = urllib2.Request(self.action, urlencode(self.items()))
            req.add_header("Content-Type", self.enctype)
            return req
        else:
            raise ValueError, "Unknown method '%s'" % method

    def click(self, name=None, coord=(1,1), items=False):
        """Return request that would result from clicking on a control.

        Will click on the first clickable control, with the indicated name, if
        specified.  Only some control types (INPUT/SUBMIT & BUTTON/SUBMIT
        buttons and IMAGEs) can be click()ed.  If no name is specified and
        there are no clickable controls, a request will be returned for the
        form in its current state.  You can optionally specify a coordinate to
        click at, which only makes a difference if you clicked on an image.

        ValueError is raised if a name is specified but no matching control is
        found.  ValueError is raised if the HTMLForm has method "POST" and an
        enctype attribute that is not "application/x-www-form-urlencoded".

        The request object is a urllib2.Request instance.

        If items is true, returns a list of (key, value) pairs instead of a
        request object.  You can pass this list to ClientForm.urlencode.  This
        is usually only useful if you're using urllib rather than urllib2 or
        ClientCookie.  It may also be useful if you want to manually tweak the
        items, but this should not be necessary.

        """
        # try to find first submit button to activate
        for control in self.controls:
            if not hasattr(control, "click"): continue
            if name is not None and control.name != name: continue
            return control.click(self, coord, items)
        if name is not None:
            raise ValueError, "No clickable control with name %s" % name
        # no clickable buttons, but no button was explicitly requested by name,
        # so return current state
        if items:
            return self.items()
        else:
            return self.make_request()

    def __repr__(self):
        header = "%s %s %s" % (self.method, self.action, self.enctype)
        rep = [header]
        for control in self.controls:
            rep.append("  %r" % control)
        return "<%s>" % string.join(rep, "\n")


class Control:
    """An HTMLForm contains a sequence of Controls.

    The Controls in an HTMLForm can be got at via the HTMLForm.find_control
    method or the HTMLForm.controls attribute.

    A Control is only properly initialised after the fixup method has been
    called.  In fact, this is only strictly necessary for ListControl
    instances.  This is necessary because ListControls are built up from
    length-1 ListControls, and their initial value(s) can only be known after
    the sequence is complete.

    It is always allowable to assign a string to the value attribute (unless
    the control is disabled or should not be assigned to for some other
    reason).  Derived classes (in particular, ListControl instances having a
    true multiple attribute) may also allow assignment of sequences.

    If the disabled attribute is true, this represents the state typically
    represented by browsers by `greying out' a control.  If the disabled
    attribute is true, the Control will raise AttributeError if an attempt is
    made to change its value.  In addition, the control will not be considered
    `successful' as defined by the W3C HTML 4 standard -- ie. its (name, value)
    pair will not be included in the output of the items method.  To enable a
    control, set the disabled attribute to a false value.

    Public attributes:

    type: see the keys of the TYPE2CLASS dictionary in this module for the
     allowable values (readonly)
    name: name of control (readonly)
    value: current value of control (subclassess may allow a single value, a
     sequence of values, or either)
    disabled: disabled state

    """
    def __init__(self, type, attrs):
        """
        type: string describing type of control (see the keys of the TYPE2CLASS
         dictionary in this module for the allowable values)
        attrs: HTML attributes of control's HTML element

        """
        self.__dict__["type"] = string.lower(type)
        self.__dict__["name"] = attrs.get("name")
        self._value = attrs.get("value")
        self.disabled = attrs.has_key("disabled")
        # This dictionary of the remaining attributes is only accessible from
        # derived classes, because the method of access must differ between
        # ScalarControl and ListControl.
        a = self._attrs = attrs.copy()
        if a.has_key("type"): del a["type"]
        if self.name is not None: del a["name"]
        if self._value is not None: del a["value"]
        if self.disabled: del a["disabled"]
        self._clicked = False

    def __setattr__(self, name, value):
        if name == "value":
            if self.disabled:
                raise AttributeError, "control '%s' is disabled" % self.name
            self.__dict__["_value"] = value
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if name == "value":
            return self.__dict__["_value"]
        else:
            raise AttributeError, "%s instance has no attribute '%s'" % (
                self.__class__.__name__, name)

    def add_to_form(self, form):
        form.controls.append(self)

    def fixup(self):
        pass

    def possible_values(self):
        """Return a list of all values that the control can take.

        For controls that do not have a prescribed set of values, this returns
        an empty list.

        """
        return []

    def items(self):
        """Return list of (key, value) pairs suitable for passing to urlencode.

        Note that Python 2.0's urllib.urlencode is slightly broken: it only
        accepts a mapping, not a sequence of pairs, as an argument.  This
        messes up any ordering in the argument.  Use ClientForm.urlencode
        instead.

        """
        name = self.name
        value = self.value
        if name is None or value is None or self.disabled:
            return []
        return [(name, value)]

    def __repr__(self):
        name = self.name
        value = self.value
        if name is None: name = "<None>"
        if value is None: value = "<None>"
        if self.disabled: dis = " (disabled)"
        else: dis = ""
        return "<%s(%s=%s)%s>" % (self.__class__.__name__, name, value, dis)


#---------------------------------------------------
class ScalarControl(Control):
    """Non-sequence control, having a single value.

    Additional public attributes:

    attrs: dictionary mapping the names of HTML attributes of the control
     to their values; only HTML attributes not accessible via the
     ScalarControl interface are found in the dictionary (so, for example,
     "name", "type" and "value" are never keys of the dictionary, but
     "alt" may be)

    """
    def __init__(self, type, attrs):
        Control.__init__(self, type, attrs)
        self.attrs = self._attrs
        del self._attrs


#---------------------------------------------------
class TextControl(ScalarControl):
    """Textual input control.

    Covers:

    INPUT/TEXT
    INPUT/PASSWORD
    INPUT/FILE
    INPUT/HIDDEN
    TEXTAREA

    If the readonly attribute is true, the Control will raise AttributeError if
    an attempt is made to change its value.  To make a control writable, set
    the readonly attribute to a false value.

    Additional public attributes:

    readonly: readonly state

    """
    def __init__(self, type, attrs):
        ScalarControl.__init__(self, type, attrs)
        self.readonly = attrs.has_key("readonly")
        if self.readonly: del self.attrs["readonly"]
        if self.type == "hidden": self.readonly = True

    def __setattr__(self, name, value):
        if name == "value":
            if self.readonly:
                raise AttributeError, "control '%s' is readonly" % self.name
            if self.disabled:
                raise AttributeError, "control '%s' is disabled" % self.name
            self.__dict__["_value"] = value
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value

    def __repr__(self):
        name = self.name
        value = self.value
        if name is None: name = "<None>"
        if value is None: value = "<None>"

        infos = []
        if self.disabled: infos.append("disabled")
        if self.readonly: infos.append("readonly")
        info = string.join(infos, ", ")
        if info: info = " (%s)" % info

        return "<%s(%s=%s)%s>" % (self.__class__.__name__, name, value, info)


#---------------------------------------------------
class IgnoreControl(ScalarControl):
    """Control that we're not interested in.

    Covers:

    INPUT/RESET
    BUTTON/RESET
    BUTTON/BUTTON

    These controls are always unsuccessful, in the terminology of HTML 4 --
    ie., they never require any information to be returned to the server.

    """
    def __setattr__(self, name, value):
        if name == "value":
            raise AttributeError, \
                  "control '%s' is ignored, hence read-only" % self.name
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value


#---------------------------------------------------
# XXX single and multiple select ListControls should really be subclasses
class ListControl(Control):
    """Control representing a sequence of items.

    Covers:

    SELECT/OPTION
    INPUT/RADIO
    INPUT/CHECKBOX

    May always assign a string to value attribute (unless control is disabled).
    May assign a sequence to value attribute if controls accepts multiple
    values (SELECT with the "multiple" HTML attribute, and CHECKBOX).

    ListControls are built up by the parser from their component items by
    creating one ListControl per item, consolidating them into a single master
    ListControls held by the HTMLForm:

    -User calls form.new_control(...)
    -Form creates Control, and calls control.add_to_form(form).
    -Control looks for an Control with the same name in the form, and if it
    finds one, merges itself with that control by calling
    control.merge_control(self).  The first Control added to the form, of a
    particular name, is the only one that survives in the form.
    -Form calls control.fixup for all its controls.  ListControls in the form
    know they can now safely pick their default values.

    To create a ListControl without an HTMLForm, use control.merge_control.

    """

    def __init__(self, type, attrs, select_default=False):
        """
        select_default: for multiple-selection SELECT controls, pick the first item
         as the default if no 'selected' HTML attribute is present
        """
        Control.__init__(self, type, attrs)
        self._select_default = select_default
        self._attrs_list = []
        # Some controls can have their default set only after all items are
        # known.  If so, self._value_is_set is false, and the self.fixup
        # method, called after all items have been added, sets the default.
        self._value_is_set = False
        if self.type == "checkbox":
            self.multiple = True
            value = attrs.get("value", "on")
            self._menu = [value]
            checked = attrs.has_key("checked")
            if checked: del self._attrs["checked"]
            self._selected = [checked]
            self._value_is_set = True
        else:
            assert self.type in ("radio", "select")
            value = attrs["value"]
            self._menu = [value]
            checked = (attrs.has_key("checked") or attrs.has_key("selected"))
            if checked:
                self._value_is_set = True
                if attrs.has_key("checked"): del self._attrs["checked"]
                if attrs.has_key("selected"): del self._attrs["selected"]
            if attrs.has_key("multiple"):
                assert self.type == "select"
                self.multiple = True
                del self._attrs["multiple"]
                self._selected = [checked]
            else:
                self.multiple = False
                if checked:
                    self._selected = value
                else:
                    self._selected = None
        self._attrs_list.append(self._attrs)
        del self._attrs

    def toggle(self, name):
        """Toggle named value."""
        if self.disabled:
            raise ValueError, "control '%s' is disabled" % self.name
        if not self.multiple:
            raise TypeError, \
                  "control %s does not have toggle-able values" % name

        for i in range(len(self._menu)):
            item_name = self._menu[i]
            if name != item_name:
                continue
            self._selected[i] = not self._selected[i]
            return
        raise ValueError, "item not found in control"

    def find_item(self, name=None, nr=0):
        """Return the dictionary of attributes for a single item in the control.

        The items we're talking about here represent the possible values of a
        list control.  The HTML elements that describe these items are: OPTION
        for SELECT controls, INPUT for the rest.  These elements have HTML
        attributes that you may occasionally want to know about -- for example,
        the "alt" HTML attribute gives a text string describing the item
        (graphical browsers usually display this as a tooltip).

        The dictionary of HTML attributes is identical to that of the attrs
        attribute of ScalarInput, as described in ScalarInput.__doc__.

        At least one of the arguments must be defined.  If no matching item
        is found, None is returned.

        If name is specified, then the item must have the indicated name (note
        that this name is usually specified as the value of an HTML attribute
        named "value").  nr is the sequence number of the item with the
        indicated name (where 0 is the first).  Note that item 0 is the first
        item matching the name argument(if supplied); it is not the first
        item in the control.

        """
        if (name is None) and (nr is None):
            raise ValueError, \
                  "At least one of name and nr args must be supplied"
        for i in range(len(self._menu)):
            item_name = self._menu[i]
            if name != item_name:
                continue
            if nr:
                nr = nr - 1
                continue
            return self._attrs_list[i]
        return None

    def add_to_form(self, form):
        control = form.find_control(self.name, self.type)
        if control is None:
            Control.add_to_form(self, form)
        else:
            control.merge_control(self)

    def merge_control(self, control):
        assert isinstance(control, self.__class__)
        # merge menus
        self._menu.extend(control._menu)
        self._attrs_list.extend(control._attrs_list)
        if control.multiple:
            self._selected.extend(control._selected)
        else:
            if control._value_is_set:
                self._selected = control._selected
        if control._value_is_set:
            self._value_is_set = True

    def fixup(self):
        """
        ListControls are built up from component list items (which are also
        ListControls) during parsing.  This method should be called after all
        items have been added.  See ListControl.__doc__ for the reason this is
        required.

        """
        # Default selection where no item was indicated as being selected by
        # the HTML:

        # CHECKBOX:
        #  Nothing should be selected.
        # SELECT/single, SELECT/multiple and RADIO:
        #  RFC 1866 (HTML 2.0): says first item should be selected.
        #  W3C HTML 4.01 Specification: says that client behaviour is
        #   undefined in this case.  For RADIO, exactly one must be selected,
        #   though which one is undefined.
        #  Both Netscape and Microsoft Internet Explorer (IE) choose first
        #   item for radio and SELECT/single.  However, both IE and Netscape
        #   leave all items unselected for SELECT/multiple.

        # Since both Netscape and IE all choose the first item for
        # SELECT/single and RADIO, we do the same.  OTOH, both Netscape and IE
        # leave SELECT/multiple with nothing selected, in violation of RFC 1866
        # (but not in violation of the W3C HTML 4 standard).  We follow RFC
        # 1866 if the select_default attribute is set, and Netscape and IE
        # otherwise.

        if not self._value_is_set:
            # No item explicitly selected.
            assert self.type in ("select", "radio")
            if self.multiple:
                assert self.type == "select"
                if self._select_default:
                    self._selected[0] = True
            else:
                if self._selected is None:
                    self._selected = self._menu[0]
            self._value_is_set = True

    def __getattr__(self, name):
        if name == "value":
            menu = self._menu
            if self.multiple:
                values = []
                for i in range(len(menu)):
                    if self._selected[i]: values.append(menu[i])
                return copy(values)
            else:
                return self._selected
        else:
            raise AttributeError, "%s instance has no attribute '%s'" % (
                self.__class__.__name__, name)

    def __setattr__(self, name, value):
        if name == "value":
            if self.disabled:
                raise AttributeError, "control '%s' is disabled" % self.name

            if not self.multiple:
                if not isstringlike(value):
                    raise AttributeError, \
                          "Single selection list, must set a string"
                try:
                    i = self._menu.index(value)
                except ValueError:
                    raise AttributeError, "Illegal value '%s'" % value
                self._selected = value
                return

            if isstringlike(value):
                value = [value]
            selected = [0]*len(self._selected)
            menu = self._menu

            for v in value:
                found = False
                for i in range(len(menu)):
                    item_name = menu[i]
                    if v == item_name:
                        selected[i] = True
                        found = True
                        break
                if not found:
                    raise AttributeError, "Illegal value '%s'" % v
            self._selected = selected
        elif name in ("name", "type"):
            raise AttributeError, "%s attribute is readonly" % name
        else:
            self.__dict__[name] = value

    def possible_values(self):
        return copy(self._menu)

    def items(self):
        if self.disabled:
            return []

        if not self.multiple:
            return Control.items(self)

        control_name = self.name  # the name attribute
        items = []
        for i in range(len(self._menu)):
            item_name = self._menu[i]  # the value attribute
            if self._selected[i]:
                items.append((control_name, item_name))
        return items

    def __repr__(self):
        menu = self._menu
        display = []
        for i in range(len(menu)):
            item_name = menu[i]
            if self.multiple:
                if self._selected[i]:
                    item_name = "*"+item_name
            else:
                if self._selected == item_name:
                    item_name = "*"+item_name
            display.append(item_name)

        if self.disabled: dis = " (disabled)"
        else: dis = ""
        return "<%s(%s, %s=[%s])%s>" % (self.__class__.__name__,
                                        self.type,
                                        self.name,
                                        string.join(display, ", "),
                                        dis)


#---------------------------------------------------
class SubmitControl(ScalarControl):
    """
    Covers:

    INPUT/SUBMIT
    BUTTON/SUBMIT

    """
    def click(self, form, coord=(1,1), items=False):
        """Return a request object simulating a click on the control.

        Only some control types (submit buttons and images) can be click()ed.

        """
        self._clicked = coord
        if items:
            r = form.items()
        else:
            r = form.make_request()
        self._clicked = False
        return r

    def items(self):
        if self.disabled or not self._clicked:
            return []
        return Control.items(self)


#---------------------------------------------------
class ImageControl(SubmitControl):
    """
    Covers:

    INPUT/IMAGE

    """
    def items(self):
        clicked = self._clicked
        if self.disabled or not self._clicked:
            return []
        name = self.name
        if name is None: return []
        return [("%s.x" % name, clicked[0]),
                ("%s.y" % name, clicked[1])]


TYPE2CLASS = {
    "text": TextControl,
    "password": TextControl,
    "hidden": TextControl,
    "textarea": TextControl,
    "isindex": TextControl,

    #"file": FileControl,

    "reset": IgnoreControl,
    "button": IgnoreControl,

    "submit": SubmitControl,
    "image": ImageControl,

    "radio": ListControl,
    "checkbox": ListControl,
    "select": ListControl,
    }
